# -*- coding: utf-8; tab-width: 4; indent-tabs-mode: nil -*-
#
# Copyright (C) 2009 - Jeremy AustinBardo <ausimage@ubuntu.com>;
#   Adam J. Lincoln <adamjlincoln@gmail.com>; Richard Harding <rharding@raken>;
#   Neil Shepperd <nshepperd@gmail.com>
# Copyright (C) 2006 - Florian Ludwig <dino@phidev.info>;
#   Franz Pletz <fpletz@franz-pletz.org>
#
# Apt-zeroconf is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# Apt-zeroconf is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with apt-zeroconf; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
"""
ZeroConf Service Module for apt-zeroconf. Track all local hosts broadcasting
availablity of personal cache service.

@requires: logging
@requires: threading
@requires: urllib.urlopen
@requires: avahi
@requires: dbus
@requires: dbus.glib
@requires: dbus
@requires: aptzeroconf.util.synchronized
@requires: aptzeroconf.config.config
@requires: gobject
@requires: aptzeroconf.util.make_url
@requires: aptzeroconf.http.AZCHTTPHandler
@requires: aptzeroconf.apt.pkglist
@requires: random
@requires: aptzeroconf.util.prune_epoch
@requires: apt_pkg.VersionCompare
"""
import logging
import threading
from urllib import urlopen
import avahi
import dbus, dbus.glib
from aptzeroconf.util import synchronized


class AZCAvahiService(threading.Thread):
    """
    Create an AZC cache service instance.

    @requires: threading
    @requires: avahi
    @requires: dbus
    """
    lock = threading.Lock()
    stype = '_apt._tcp'
    domain = ''
    host = ''
    txt = []

    def __init__(self):
        """
        Intialize an AZC cache service instance. Create hosts dictionary for
        any non-local hosts discovered.

        @requires: aptzeroconf.config.config
        """
        from aptzeroconf.config import config
        threading.Thread.__init__(self)
        self.group = None
        self.n_rename = 0

        self.name = "apt-zeroconf repository on %s" % config['hostname']
        self.port = config['http_port']

        hosts.set_service(self)
        self.hosts = hosts

    def _remove_service(self):
        """
        Remove an AZC cache service instance.
        """
        if self.group:
            self.group.Reset()

    def _add_service(self):
        """
        Add an AZC cache service instance.
        """
        if not self.group:
            self.group = dbus.Interface(self.bus.get_object(avahi.DBUS_NAME,
                self.server.EntryGroupNew()), avahi.DBUS_INTERFACE_ENTRY_GROUP)
            self.group.connect_to_signal('StateChanged',
                self._entry_group_state_changed)
        else:
            return

        assert self.group.IsEmpty()
        logging.debug('Creating service \'%s\' (%s)' % (self.name, self.stype))

        try:
            self.group.AddService(avahi.IF_UNSPEC, avahi.PROTO_UNSPEC, 0,
                self.name, self.stype, self.domain, self.host,
                dbus.UInt16(self.port), avahi.string_array_to_txt_array(self.txt))
        except dbus.DBusException:
            self.name = self.name + ' 1'
            logging.debug('Creating service \'%s\' (%s)' % (self.name, self.stype))
            self.group.AddService(avahi.IF_UNSPEC, avahi.PROTO_UNSPEC, 0,
                self.name, self.stype, self.domain, self.host,
                dbus.UInt16(self.port), avahi.string_array_to_txt_array(self.txt))
        self.group.Commit()

    @synchronized(lock)
    def _entry_group_state_changed(self, state, wtf):
        """
        Thread locked handler for an AZC cache service group state change. Will
        alter the host name when a namespace collision occur.

        @keyword state: Group state change
        @type state: basestring
        @keyword wtf:
        @type wtf:

        """
        if state == avahi.ENTRY_GROUP_ESTABLISHED:
            logging.debug('Service established')

        elif state == avahi.ENTRY_GROUP_COLLISION:
            self.n_rename += 1

            if self.n_rename >= 12:
                logging.error('No suitable service name found after %i '
                    'retries, exiting' % self.n_rename)
                self.stop()

            else:
                self.name = self.server.GetAlternativeServiceName(self.name)
                logging.warning('Service name collision, changing name to '
                    '\'%s\'' % self.name)
                self._remove_service()
                self._add_service()

    @synchronized(lock)
    def _server_state_changed(self, state, *args):
        """
        Thread locked handler for AZC cache service server state change.

        @keyword state: Server state change
        @type state: basestring
        """
        if state == avahi.SERVER_COLLISION:
            logging.warning('Server name collision')
            self._remove_service()
        elif state == avahi.SERVER_RUNNING:
            self._add_service()

    @synchronized(lock)
    def _name_owner_changed(self, name, old, new):
        """
        Thread locked handler for AZC cache service owner state change.

        @keyword name: Avahi service name
        @type name: basestring
        @keyword new: New Avahi service name
        @type new: basestring
        @keyword old: Old Avahi service name
        @type old: basestring
        """
        if name == avahi.DBUS_NAME:

            if not new:
                logging.info('Avahi server appeared')
                self.lock.release()
                self._avahi_connect()
                self.lock.acquire()

            elif not old:
                logging.info('Avahi server disappeard')
                self.group = None
                self.hosts.clear()

    def _dbus_connect(self):
        """
        Handle an AZC cache service instance connection to dbus.
        """
        self.dbus = dbus.Interface(
                self.bus.get_object('org.freedesktop.DBus',
                    '/org/freedesktop/DBus'), 'org.freedesktop.DBus')
        self.dbus.connect_to_signal('NameOwnerChanged', self._name_owner_changed)

    def _avahi_connect(self):
        """
        Handle an AZC cache service instance connection.
        """
        try:
            self.server = dbus.Interface(
                self.bus.get_object(avahi.DBUS_NAME, avahi.DBUS_PATH_SERVER),
                avahi.DBUS_INTERFACE_SERVER)
        except dbus.DBusException:
            return

        self.server.connect_to_signal('StateChanged', self._server_state_changed)
        self._server_state_changed(self.server.GetState())

        b = dbus.Interface(self.bus.get_object(avahi.DBUS_NAME,
            self.server.ServiceBrowserNew(avahi.IF_UNSPEC, avahi.PROTO_UNSPEC,
                self.stype, 'local', dbus.UInt32(0))),
            avahi.DBUS_INTERFACE_SERVICE_BROWSER)
        b.connect_to_signal('ItemNew', self.hosts.add)
        b.connect_to_signal('ItemRemove', self.hosts.remove)

    def run(self):
        """
        Setup an AZC cache service instance for use.
        @requires: dbus
        @requires: avahi
        """
        self.bus = dbus.SystemBus()
        self._dbus_connect()
        self._avahi_connect()

        self.loop()

        if not self.group is None:
            self.group.Free()

    def loop(self):
        """
        Start an AZC cache service instance for use.
        @requires: gobject
        """
        import gobject
        self.ml = gobject.MainLoop()
        gobject.threads_init()
        self.ml.run()

    def stop(self):
        """
        Stop an AZC cache service instance for use.
        """
        self.ml.quit()


class Hosts(dict):
    """
    Store non-local AZC cache service hosts in a dictionary. Provide mechanism
    to fetch package lists from these hosts as well

    @requires: threading
    """
    lock = threading.Lock()
    service = None

    @synchronized(lock)
    def add(self, interface, protocol, name, stype, domain, flags):
        """
        Add non-local AZC cache service hosts. Call an AZCAvahiService class
        instance of self.service.server.ResolveService which replies to set the
        host values with self.service_resolved or errors with logging.error

        @keyword interface: Avahi service interface [avahi.IF_UNSPEC]
        @type interface: int
        @keyword protocol: Avahi service protocol value[avahi.PROTO_UNSPEC]
        @type protocol: int
        @keyword name: Avahi service description
        @type name: basestring
        @keyword stype: Avahi service type ['_apt._tcp']
        @type stype: basestring
        @keyword domain: Avahi service domain ['local']
        @type domain: basestring
        @keyword flags:
        @type flags:
        @requires: avahi
        """
        if not self.service:
            raise AttributeError('Service is not set. Use set_service() first.')

        if self.service.name == name:
            # We found ourselves
            return

        logging.info('Found service \'%s\'' % name)
        self.service.server.ResolveService(int(interface), int(protocol),
            name, stype, domain, avahi.PROTO_UNSPEC, dbus.UInt32(0),
            reply_handler=self.service_resolved, error_handler=logging.error)

    @synchronized(lock)
    def service_resolved(self, interface, protocol, name, stype, domain, host,
        aprotocol, address, port, txt, flags):
        """
        Store non-local AZC cache service host values. Acts a reply handler for
        self.service.server.ResolveService.

        @keyword interface: Avahi service interface [avahi.IF_UNSPEC]
        @type interface: int
        @keyword protocol: Avahi service protocol value[avahi.PROTO_UNSPEC]
        @type protocol: int
        @keyword name: Avahi service description
        @type name: basestring
        @keyword stype: Avahi service type ['_apt._tcp']
        @type stype: basestring
        @keyword domain: Avahi service domain ['local']
        @type domain: basestring
        @keyword host:
        @type host:
        @keyword aprotocol:
        @type aprotocol:
        @keyword address: Avahi cache service host address
        @type address: basestring
        @keyword port: Avahi cache service host port
        @type port: int
        @keyword txt: Avahi cache service host descriptor
        @type txt: list
        @keyword flags:
        @type flags:
        @requires: aptzeroconf.util.make_url
        """
        from aptzeroconf.util import make_url
        if address.startswith('fe80:'):
            logging.info('%s -> %s: Local Link IPv6 addresses not supported!'
                % (host, address))
            return

        url = make_url(address, port, protocol)
        self['%s:%s:%s' % (name, interface, domain)] = {
                'host': host, 'address': address, 'port': port, 'txt': txt,
                'proto': protocol, 'url': url,
                'pkglist': self._fetch_list(url)}
        logging.debug('New host \'%s\' (%s)' %
                (name, url))

    def _fetch_list(self, url):
        """
        Fetch non-local AZC Cache service host cache listings.

        @param url:
        @type url:
        @requires: aptzeroconf.http.AZCHTTPHandler
        @requires: aptzeroconf.apt.pkglist
        @todo: should we change the list parsing to something better?
        """
        from aptzeroconf.http import PROTO_VER
        from aptzeroconf.apt import LIST_SEP
        data = urlopen('http://%s/list' % url).read().split(LIST_SEP)
        if data[0] != PROTO_VER:
            logging.debug('Host %s uses incompatible format ' % url +
                'version (%s) of package list!' % data[0])
            return []

        return data[1:]

    @synchronized(lock)
    def update_list(self, address):
        """
        Update non-local AZC cache service host cache listings.

        @keyword address:
        @type address:
        """
        for data in self.itervalues():
            if data['address'] == address:
                data['pkglist'] = self._fetch_list(data['url'])

    @synchronized(lock)
    def remove(self, interface, protocol, name, stype, domain, flags):
        """
        Remove non-local AZC cache service hosts.
        @keyword interface: Avahi service interface [avahi.IF_UNSPEC]
        @type interface: int
        @keyword protocol: Avahi service protocol value[avahi.PROTO_UNSPEC]
        @type protocol: int
        @keyword name: Avahi service description
        @type name: basestring
        @keyword stype: Avahi service type ['_apt._tcp']
        @type stype: basestring
        @keyword domain: Avahi service domain ['local']
        @type domain: basestring
        @keyword flags:
        @type flags:
        """
        logging.info('Service \'%s\' disappeared' % name)
        try:
            del self['%s:%s:%s' % (name, interface, domain)]
        except KeyError:
            pass

    @synchronized(lock)
    def set_service(self, service):
        """
        Set non-local AZC cache service object.

        @keyword address:
        @type address:
        """
        self.service = service

    @synchronized(lock)
    def get_urls_for_file(self, f):
        """
        Find file in available AZC cache services.

        @keyword f:filename to find
        @type f:
        @requires: random
        @requires: aptzeroconf.util.prune_epoch
        @requires: apt_pkg.VersionCompare
        """
        import random
        from aptzeroconf.util import prune_epoch
        from apt_pkg import VersionCompare
        hosts = []
        for data in self.itervalues():
            if not '%3a' in f:
                # find the file with the latest epoch
                finds = [pkg for pkg in data['pkglist']
                    if prune_epoch(pkg) == f]

                if not finds:
                    continue

                finds.sort(cmp=VersionCompare)
                f = finds[-1]
                found = True

            elif f in data['pkglist']:
                found = True

            if found:
                hosts.append(data['url'])

        random.shuffle(hosts)
        return ['http://%s/debian/%s' % (h, f) for h in hosts]

    @synchronized(lock)
    def propagate_update(self):
        """
        Propagate Updates. Used by apt module. Call each known AZC cache service
        by accessing their update url to notify of this host's updates.
        """
        for data in self.itervalues():
            threading.Thread(target=urlopen('http://%s/list/update' %
                data['url']).read).start()

    @synchronized(lock)
    def clear(self):
        """
        Propagate Updates. Used by apt module.
        """
        dict.clear(self)


hosts = Hosts()

