# -*- coding: utf-8; tab-width: 4; indent-tabs-mode: nil -*-
#
# Copyright (C) 2009 - Jeremy AustinBardo <ausimage@ubuntu.com>;
#   Adam J. Lincoln <adamjlincoln@gmail.com>; Craig Maloney <craig@decafbad.net>;
#   Neil Shepperd <nshepperd@gmail.com>
# Copyright (C) 2006 - Florian Ludwig <dino@phidev.info>;
#    Franz Pletz <fpletz@franz-pletz.org>
#
# Apt-zeroconf is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# Apt-zeroconf is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with apt-zeroconf; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
"""
HTTP Service Module for apt-zeroconf. Translates path into either a proxied,
redirected, or page url.

@requires: logging
@requires: SocketServer.[ThreadingTCPServer|BaseServer]
@requires: SimpleHTTPServer.SimpleHTTPRequestHandler
@requires: aptzeroconf.config.[config|cache]
@requires: socket
@requires: urllib2.HTTPError
"""
import logging, socket
import os.path
from SocketServer import ThreadingTCPServer, BaseServer
from SimpleHTTPServer import SimpleHTTPRequestHandler

from aptzeroconf.config import config, cache

PROTO_VER = '0.2'


class HttpServer(ThreadingTCPServer):
    """
    HTTP Proxy server.
    """
    def __init__(self):
        BaseServer.__init__(self, ('', config['http_port']), AZCHTTPHandler)
        self.socket = socket.socket(socket.AF_INET6, self.socket_type)
        self.socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self.server_bind()
        self.server_activate()


class AZCHTTPHandler(SimpleHTTPRequestHandler):
    """
    HTTP Proxy handler.
    """
    HOSTNAME = "http://%s.local" % config['hostname']

    def do_POST(self):
        """
        Respond to a POST request by proxy the url to preferably local content,
        otherwise internet content. This method recognizes three states of POST.

            - Proxy Traffic
            - Redirect Traffic

        @requires: urllib2.HTTPError
        @note: Required for apt-listbugs to work properly.
        """
        from urllib2 import HTTPError

        logging.info("[%s]: %s (POST)" % (self.client_address[0], self.path))
        self.url = self.path
        errcode = self.__proxy() #: Check for valid paths and set if valid.

        #{ Handle unauthorized proxy urls, when config['restrict] is set
        if errcode == 403:
            self.send_error(403)

        #{ Handle proxy and redirect urls
        elif errcode == 200:
            try:
                self.__proxy_content()
            except HTTPError as error:
                self.send_error(error.code)

        #{ Handle any other paths that occur
        else:
            self.send_error(404)

    def do_GET(self):
        """
        Wrapper for real _do_GET() method to trap SocketErrors that may occur.

        @requires: socket.error
        """
        from socket import error as SocketError
        try:
            self._do_GET()
        except SocketError:
            logging.info("[%s]: Connection reset by peer" % (
                self.client_address[0]))

    def _do_GET(self):
        """
        Respond to a GET request by proxy the url to preferably local content,
        otherwise internet content. This method recognizes three states of GET.

            - Proxy Traffic
            - Redirect Traffic
            - Local Traffic

        @requires: os.path.exists
        @requires: aptzeroconf.apt.pkglist
        @requires: aptzeroconf.avaphi.hosts
        @requires: aptzeroconf.stats.[www_recv|lan_recv]
        """
        logging.info('[%s]: %s (GET)' % (self.client_address[0], self.path))
        self.url = self.path

        errcode = self.__proxy() #: Check for valid paths and set if valid.

        #{ Handle unauthorized proxy urls, when config['restrict] is set
        if errcode == 403:
            self.send_error(403)

        #{ Handle proxy and redirect urls
        elif errcode == 200:
            self.fetch_file()

        #{ Handle any other paths that occur
        else:

            from aptzeroconf.__init__ import __version__
            def pagepath(filename):
                if os.path.isfile("pages/" + filename):
                    return "pages/" + filename
                else:
                    return "/usr/share/pages/" + filename

            httpCode = 200
            httpType = "text/html"
            pagePath = self.path
            pageValues = {"version": __version__,}

            logging.info("[%s]: Website Content" % self.client_address[0])

            #{ Handle /list path and its children
            if self.path.startswith("/list"):
                from aptzeroconf.apt import pkglist

                httpType = "text/plain"
                pageData = open(pagepath("list.html"),"r")

                if self.path.startswith("/list/update"):
                    from aptzeroconf.avaphi import hosts
                    logging.debug("[Local Command] Process package updates")
                    hosts.update_list(self.client_address[0])
                    pageValues['status'] = "Updated local cache package list.\n"

                else:
                    logging.debug("[Local Page] Cache package list")
                    pageValues['status'] = ""

                pageValues['packages'] = "\n".join([PROTO_VER,
                    str(pkglist)])

            #{ Handle /host path and its children
            elif self.path.startswith("/host"):
                from aptzeroconf.avaphi import hosts

                pageData = open(pagepath("host.html"),"r")

                if self.path.startswith("/host/file"):
                    from urllib2 import unquote
                    # This is hack. to use file's child must be a full pkg name.
                    logging.debug("[Local Command] Show package locations")
                    pageValues['cachehosts'] = "\n".join(
                        [("<li>%s</li>" % service)
                        for service in self._get_pkgurls() if service])
                    pageValues['package'] = "with %s" % (
                        unquote(self.path.rsplit("/",1)[1]))

                else:
                    logging.debug("[Local Page] Cache host list")
                    pageValues ['cachehosts'] = "\n".join(
                        [("<li>%s</li>" % service)
                        for service in hosts if service])
                    pageValues['package'] = ""

            #{ Handle /stats path and its children
            elif self.splitUrl.path.startswith('/stats'):
                from aptzeroconf.stats import www_recv, lan_recv

                pageData = open(pagepath("stats.html"),"r")
                logging.debug('[Local Page]: Cache traffic use details')

                if config['stats'].lower() != 'off':
                    total_recv = int(www_recv + lan_recv)
                    pageValues ['status'] = ("Current cache use for this host.")
                    pageValues ['usewan'] = (
                        float(www_recv / total_recv)
                        if total_recv !=0 else 0)
                    pageValues ['uselan'] = (
                        float(lan_recv / total_recv)
                        if total_recv !=0 else 0)

                else:
                    pageValues ['status'] = (
                        "No current use details available for this host. "
                        "Please, enable stats in the configuration to use.")
                    pageValues ['usewan'] = 0
                    pageValues ['uselan'] = 0

            #{ Handle anyother path
            else:
                pageData = open(pagepath("index.html"),"r")
                pageValues ['status'] = ""

            #{ Compose and transmit page based on current path value.
            self.__comp_Page(httpCode, httpType, pageData, pageValues)

    def fetch_file(self):
        """
        Handle location of retreived proxy content from either URL or local
        cache. Currently only files ending with '.deb' are checked locally.

        @requires: urllib2.[unquote|HTTPError]
        """
        from urllib2 import unquote, HTTPError

        orig_path = self.path

        logging.info("[%s]: Fallback url: %s" % (self.client_address[0],
            orig_path))

        #{ Handle location of retreived debian package files
        if unquote(self.path).endswith(".deb"):
            for curr_path in self._get_pkgurls(orig_path):
                logging.info("[%s]: Current url: %s" % (self.client_address[0],
                    curr_path))
                self.url = curr_path
                try:
                    self.__proxy_content(orig_path=orig_path)
                except HTTPError as error:
                    continue
                break
            else:
                self.send_error(404)
                return

            logging.info("[%s]: Actual url is %s" % (self.client_address[0],
                curr_path))

        #{ Handle location of other retreived files
        else:
            try:
                self.__proxy_content(orig_path=orig_path)
            except HTTPError as error:
                self.send_error(error.code)

#    def _from_localhost(self):
#        """
#        Determine if current client is a local address.
#        """
#        return self.client_address[0] in ['127.0.0.1', '::1',
#           '::ffff:127.0.0.1', self.HOSTNAME]

    def _get_pkgurls(self, orig_path=None):
        """
        Handle locating a file on the local area network by querying other
        apt-zero-conf instances. Currently only files ending with '.deb' are
        checked.

        @param orig_path: original full url to apend as last resort location
        @type orig_path: basestring
        @return: host list for package
        @rtype: list
        @requires: urllib.[quote|unquote]
        @requires: aptzeroconf.apt.pkglist
        @requires: aptzeroconf.avaphi.hosts
        """
        from urllib import quote, unquote
        from aptzeroconf.apt import pkglist
        from aptzeroconf.avaphi import hosts

        urls = []
        pkg = unquote(self.path.rsplit('/', 1)[1])

        #{ Check for package file on this host
        if pkglist.has(pkg):
            urls.append('file://%s' % quote(pkglist.get_path(pkg)))

        if self.path.startswith('/debian'):
            # this is another azc instance asking for our local file
            # so we'll stop now
            return urls

        #{ Check for package file in other lan caches
        # Not sure about limiting to just this host, if not local a request
        #if self._from_localhost():
        urls.extend(hosts.get_urls_for_file(pkg))

        #{ Last we append the original url
        urls.append(orig_path)
        return urls

    def log_request(self, code='-', size='-'):
        pass

    #{ Page Composition Methods
    def __comp_Page(self, httpCode, httpType, httpPage, httpValues):
        """
        Compose page from httpPage template while merging httpValues. Adjusts
        headers for Content-Type and Content-Length. Finally transmit the page.

        @param httpCode: Http Error Code to transmit
        @type httpCode: int
        @param httpType: Http Mime type to transmit
        @type httpType: basestring
        @param httpPage: Page template to build Http body
        @type httpPage: file object
        @param httpValues: Page values to substitue in page template
        @type httpValues: dict
        @requires: string.Template
        """
        from string import Template

        page = Template(httpPage.read())
        httpBody = page.safe_substitute(httpValues)

        self.send_response(httpCode)
        self.send_header('Content-Type', httpType)
        self.send_header('Content-Length', len(httpBody))
        self.end_headers()
        self.wfile.write(httpBody)

    #{ Proxy Content Methods
    def __proxy(self):
        """
        Sets new url if path matches a proxy or redirect. Returns
        with HTTP Error code corresponding to its availablity.

        @return: HTTP error code
        @rtype: int
        """
        logging.debug('url: ' + self.path)

        #{ Set HTTP Proxy url. Restricted or unrestricted proxy access
        if self.path.lstrip('/').startswith('http://'):
            self.url = self.path = self.path.lstrip('/')
            if self.__proxy_http():
                logging.info('    Authorized URL: %s' % (
                    self.path))
                return 200

            else:
                logging.info('    Unauthorized URL: %s' % (
                    self.path))
                return 403

        #{ /debian/${package}.deb url asked for by other azc instances
        if self.path.startswith('/debian'):
            logging.info('    Local Package: %s' % (
                    self.path.rsplit('/',1)[1]))
            return 200

        #{ Set HTTP Redirect url as set in apt-zerconf.conf [apt-cache]
        elif self.__proxy_name():
            logging.info('    URL: %s' % (self.url))
            return 200

        return 404

    def __proxy_http(self):
        """
        Check a verified proxy path to see if host is in restricted mode and
        if proxy path is in proxy list else just return True.

            - Closed Proxy: Only URLS in apt-zerconf.conf [apt-cache] values
                allowed
            - Open Proxy: Any URLS allowed

        @return: Existence of a proxy url
        @rtype: bool
        """

        if config['restrict'].lower() == 'on':
            logging.info("[%s]: Restricted Proxy" % self.client_address[0])
            if not any([any([self.path.startswith(url)
                    for url in value]) for value in cache.values()]):
                return False
        else:
            logging.info("[%s]: Unrestricted Proxy" % self.client_address[0])

        return True

    def __proxy_name(self):
        """
        Check a path to see if it is a name for a proxy url as found in
        apt-zerconf.conf [apt-cache] keys, then modify the path to be the value
        of that key.

        @return: Existence of a path redirect
        @rtype: bool
        """
        # FIXME I'm not sure if this is right... need a test for it
        for label in cache.keys():
            if self.path.startswith('/' + label):
                self.path = self.url = self.path.lstrip('/')
                logging.info("[%s]: Redirected Proxy" % (
                self.client_address[0]))
                self.url = cache[label][0] + self.path[len(label):]
#                splitUrl = list(self.splitUrl)
#                splitUrl[1] = urlsplit(cache[label][0]).netloc
#                splitUrl[2] = urljoin(urlsplit(cache[label][0]).path,
#                    self.splitUrl.path[len(label[0]):])
#                self.splitUrl = urlsplit(urlunsplit(tuple(splitUrl)))

                return True

        return False

    def __proxy_content(self, **keywords):
        """
        Handle non-local content requests. Alter some headers to reflect its
        destination url and not this local host. Add any POST query data passed.

        @keyword query:
        @type query:
        @keyword orig_path: Original url path
        @type orig_path: basetring
        @requires: urllib2.[Request|urlopen]
        @requires: __init__.__version__
        @requires: aptzeroconf.stats[www_recv|lan_recv]
        """
        from __init__ import __version__
        from urllib2 import Request, urlopen

        query = keywords.get("query", None)
        orig_path = keywords.get("orig_path", None)
        curr_path = self.url

        #{ Modify current headers for Host and User-Agent
        self.headers['User-Agent'] = 'apt-zeroconf %s (%s)' % (
            __version__, self.headers['User-Agent'])
#        self.headers['Host'] = self.splitUrl.netloc
        #{ Request data from proxy url path
        data = urlopen(Request(url=curr_path, data=query, headers=self.headers))
        headers = data.info()

        #{ Transmit content
        self.send_response(200)
        for (key, value) in headers.items():
            self.send_header(key, value)
        self.end_headers()
        self.copyfile(data, self.wfile)

        #{ Record content size to a tracker
        if (config['stats'].lower() != 'off' and
            'content-length' in self.headers):
            from aptzeroconf.stats import www_recv, lan_recv
            recv = (lan_recv if curr_path != orig_path else www_recv)
            recv.add(int(headers['content-length']))
